import collections
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from multipledispatch import Dispatcher

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.util as util

convert = Dispatcher(
    'convert',
    doc="""\
Convert `column` to the pandas dtype corresponding to `out_dtype`, where the
dtype of `column` is `in_dtype`.

Parameters
----------
in_dtype : Union[np.dtype, pandas_dtype]
    The dtype of `column`, used for dispatching
out_dtype : ibis.expr.datatypes.DataType
    The requested ibis type of the output
column : pd.Series
    The column to convert

Returns
-------
result : pd.Series
    The converted column
""",
)


class Schema:

    """An object for holding table schema information, i.e., column names and
    types.

    Parameters
    ----------
    names : Sequence[str]
        A sequence of ``str`` indicating the name of each column.
    types : Sequence[DataType]
        A sequence of :class:`ibis.expr.datatypes.DataType` objects
        representing type of each column.
    """

    __slots__ = 'names', 'types', '_name_locs'

    def __init__(self, names, types):
        if not isinstance(names, list):
            names = list(names)

        self.names = names
        self.types = list(map(dt.dtype, types))

        self._name_locs = {v: i for i, v in enumerate(self.names)}

        if len(self._name_locs) < len(self.names):
            duplicate_names = list(self.names)
            for v in self._name_locs.keys():
                duplicate_names.remove(v)
            raise com.IntegrityError(
                f'Duplicate column name(s): {duplicate_names}'
            )

    def __repr__(self):
        space = 2 + max(map(len, self.names), default=0)
        return "ibis.Schema {{{}\n}}".format(
            util.indent(
                ''.join(
                    f'\n{name.ljust(space)}{str(type)}'
                    for name, type in zip(self.names, self.types)
                ),
                2,
            )
        )

    def __hash__(self):
        return hash((type(self), tuple(self.names), tuple(self.types)))

    def __len__(self):
        return len(self.names)

    def __iter__(self):
        return iter(self.names)

    def __contains__(self, name):
        return name in self._name_locs

    def __getitem__(self, name):
        return self.types[self._name_locs[name]]

    def __getstate__(self):
        return {slot: getattr(self, slot) for slot in self.__class__.__slots__}

    def __setstate__(self, instance_dict):
        for key, value in instance_dict.items():
            setattr(self, key, value)

    def delete(self, names_to_delete):
        for name in names_to_delete:
            if name not in self:
                raise KeyError(name)

        new_names, new_types = [], []
        for name, type_ in zip(self.names, self.types):
            if name in names_to_delete:
                continue
            new_names.append(name)
            new_types.append(type_)

        return Schema(new_names, new_types)

    @classmethod
    def from_tuples(cls, values):
        if not isinstance(values, (list, tuple)):
            values = list(values)

        names, types = zip(*values) if values else ([], [])
        return Schema(names, types)

    @classmethod
    def from_dict(cls, dictionary):
        return Schema(*zip(*dictionary.items()))

    def equals(self, other, cache=None):
        return self.names == other.names and self.types == other.types

    def __eq__(self, other):
        return self.equals(other)

    def __gt__(self, other):
        return set(self.items()) > set(other.items())

    def __ge__(self, other):
        return set(self.items()) >= set(other.items())

    def append(self, schema):
        return Schema(self.names + schema.names, self.types + schema.types)

    def items(self):
        return zip(self.names, self.types)

    def name_at_position(self, i):
        """ """
        upper = len(self.names) - 1
        if not 0 <= i <= upper:
            raise ValueError(
                'Column index must be between 0 and {:d}, inclusive'.format(
                    upper
                )
            )
        return self.names[i]

    def apply_to(self, df):
        """Applies the Ibis schema to a pandas DataFrame

        Parameters
        ----------
        df : pandas.DataFrame

        Returns
        -------
        df : pandas.DataFrame

        Notes
        -----
        Mutates `df`
        """

        for column, dtype in self.items():
            pandas_dtype = dtype.to_pandas()
            col = df[column]
            col_dtype = col.dtype

            try:
                not_equal = pandas_dtype != col_dtype
            except TypeError:
                # ugh, we can't compare dtypes coming from pandas,
                # assume not equal
                not_equal = True

            if not_equal or isinstance(dtype, dt.String):
                df[column] = convert(col_dtype, dtype, col)

        return df


class HasSchema:

    """
    Base class representing a structured dataset with a well-defined
    schema.

    Base implementation is for tables that do not reference a particular
    concrete dataset or database table.
    """

    def __repr__(self):
        return f'{type(self).__name__}({repr(self.schema)})'

    def has_schema(self):
        return True

    def equals(self, other, cache=None):
        return type(self) == type(other) and self.schema.equals(
            other.schema, cache=cache
        )

    def root_tables(self):
        return [self]

    @property
    def schema(self):
        raise NotImplementedError


schema = Dispatcher('schema')
infer = Dispatcher('infer')


@schema.register(Schema)
def identity(s):
    return s


@schema.register(collections.abc.Mapping)
def schema_from_mapping(d):
    return Schema.from_dict(d)


@schema.register(collections.abc.Iterable)
def schema_from_pairs(lst):
    return Schema.from_tuples(lst)


@schema.register(collections.abc.Iterable, collections.abc.Iterable)
def schema_from_names_types(names, types):
    return Schema(names, types)


def coerce_to_tuple(
    data: Union[List, np.ndarray, pd.Series],
    output_type: dt.Struct,
    index: Optional[pd.Index] = None,
) -> Tuple:
    """Coerce the following shapes to a tuple.

    (1) A list
    (2) An np.ndarray
    (3) A Series
    """
    return tuple(data)


def coerce_to_np_array(
    data: Union[List, np.ndarray, pd.Series],
    output_type: dt.Struct,
    index: Optional[pd.Index] = None,
) -> np.ndarray:
    """Coerce the following shapes to an np.ndarray.

    (1) A list
    (2) An np.ndarray
    (3) A Series
    """
    return np.array(data)


def coerce_to_series(
    data: Union[List, np.ndarray, pd.Series],
    output_type: dt.DataType,
    original_index: Optional[pd.Index] = None,
) -> pd.Series:
    """Coerce the following shapes to a Series.

    (1) A list
    (2) An np.ndarray
    (3) A Series

    Note:
    This method does NOT always return a new Series. If a Series is
    passed in, this method will return the original object.

    Parameters
    ----------
    data : pd.Series, a list, or an np.array

    output_type : the type of the output

    original_index : Optional parameter containing the index of the output

    Returns
    -------
    pd.Series
    """
    if isinstance(data, (list, np.ndarray)):
        result = pd.Series(data)
    elif isinstance(data, pd.Series):
        result = data
    else:
        # This case is a non-vector elementwise or analytic UDF that should
        # not be coerced to a Series.
        return data
    if original_index is not None:
        result.index = original_index
    return result


def coerce_to_dataframe(
    data: Any,
    output_type: dt.Struct,
    original_index: Optional[pd.Index] = None,
) -> pd.DataFrame:
    """Coerce the following shapes to a DataFrame.

    The following shapes are allowed:
    (1) A list/tuple of Series
    (2) A list/tuple np.ndarray
    (3) A list/tuple of scalars
    (4) A Series of list/tuple
    (5) pd.DataFrame

    Note:
    This method does NOT always return a new DataFrame. If a DataFrame is
    passed in, this method will return the original object.

    Parameters
    ----------
    data : pd.DataFrame, a tuple/list of pd.Series, a tuple/list of np.array
    or a pd.Series of tuple/list

    output_type : a Struct containing the names and types of the output

    original_index : Optional parameter containing the index of the output

    Returns
    -------
    pd.DataFrame

    Examples
    --------
    >>> coerce_to_dataframe(pd.DataFrame({'a': [1, 2, 3]}), dt.Struct([('b', 'int32')]))  # noqa: E501
       b
    0  1
    1  2
    2  3
    dtype: int32
    >>> coerce_to_dataframe(pd.Series([[1, 2, 3]]), dt.Struct([('a', 'int32'), ('b', 'int32'), ('c', 'int32')]))  # noqa: E501
       a  b  c
    0  1  2  3
    dtypes: [int32, int32, int32]
    >>> coerce_to_dataframe(pd.Series([range(3), range(3)]), dt.Struct([('a', 'int32'), ('b', 'int32'), ('c', 'int32')]))  # noqa: E501
       a  b  c
    0  0  1  2
    1  0  1  2
    dtypes: [int32, int32, int32]
    >>> coerce_to_dataframe([pd.Series(x) for x in [1, 2, 3]], dt.Struct([('a', 'int32'), ('b', 'int32'), ('c', 'int32')]))  # noqa: E501
       a  b  c
    0  1  2  3
    >>>  coerce_to_dataframe([1, 2, 3], dt.Struct([('a', 'int32'), ('b', 'int32'), ('c', 'int32')]))  # noqa: E501
       a  b  c
    0  1  2  3
    dtypes: [int32, int32, int32]
    """
    if isinstance(data, pd.DataFrame):
        result = data
    elif isinstance(data, pd.Series):
        if not len(data):
            result = data.to_frame()
        else:
            num_cols = len(data.iloc[0])
            series = [data.apply(lambda t: t[i]) for i in range(num_cols)]
            result = pd.concat(series, axis=1)
    elif isinstance(data, (tuple, list, np.ndarray)):
        if isinstance(data[0], pd.Series):
            result = pd.concat(data, axis=1)
        elif isinstance(data[0], np.ndarray):
            result = pd.concat([pd.Series(v) for v in data], axis=1)
        else:
            # Promote scalar to Series
            result = pd.concat([pd.Series([v]) for v in data], axis=1)
    else:
        raise ValueError(f"Cannot coerce to DataFrame: {data}")

    result.columns = output_type.names
    if original_index is not None:
        result.index = original_index
    return result



from pathlib import Path

import pyarrow as pa
import pyarrow.parquet as pq
import pytest

import ibis
from ibis.backends.pandas.tests.conftest import TestConf as PandasTest


class TestConf(PandasTest):
    check_names = False
    supports_divide_by_zero = True
    returned_timestamp_unit = 'ns'

    @staticmethod
    def connect(data_directory: Path):
        filename = data_directory / 'functional_alltypes.parquet'
        if not filename.exists():
            pytest.skip(f'test data set {filename} not found')
        return ibis.parquet.connect(data_directory)


@pytest.fixture
def parquet(tmpdir, file_backends_data):
    # create single files
    d = tmpdir.mkdir('pq')

    for k, v in file_backends_data.items():
        f = d / f'{k}.parquet'
        table = pa.Table.from_pandas(v)
        pq.write_table(table, str(f))

    return ibis.parquet.connect(tmpdir).database()



"""Dask backend execution of struct fields and literals."""

import operator

import dask.dataframe as dd
import dask.dataframe.groupby as ddgb

import ibis.expr.operations as ops

from ..dispatch import execute_node
from .util import make_selected_obj


@execute_node.register(ops.StructField, dd.Series)
def execute_node_struct_field_series(op, data, **kwargs):
    field = op.field
    # TODO This meta is not necessarily right
    return data.map(
        operator.itemgetter(field), meta=(data.name, data.dtype)
    ).rename(field)


@execute_node.register(ops.StructField, ddgb.SeriesGroupBy)
def execute_node_struct_field_series_group_by(op, data, **kwargs):
    field = op.field
    selected_obj = make_selected_obj(data)
    return (
        selected_obj.map(operator.itemgetter(field), meta=selected_obj._meta)
        .rename(field)
        .groupby(data.index)
    )



from multipledispatch import Dispatcher

import ibis.backends.pandas.core as core_dispatch
import ibis.backends.pandas.dispatch as pandas_dispatch
from ibis.backends.dask.trace import TraceTwoLevelDispatcher

execute_node = TraceTwoLevelDispatcher('execute_node')
for types, func in pandas_dispatch.execute_node.funcs.items():
    execute_node.register(*types)(func)

execute = Dispatcher('execute')
execute.funcs.update(core_dispatch.execute.funcs)

pre_execute = Dispatcher('pre_execute')
pre_execute.funcs.update(core_dispatch.pre_execute.funcs)

execute_literal = Dispatcher('execute_literal')
execute_literal.funcs.update(core_dispatch.execute_literal.funcs)

post_execute = Dispatcher('post_execute')
post_execute.funcs.update(core_dispatch.post_execute.funcs)



import numpy as np
import pandas as pd
import pytest

ROWID_ZERO_INDEXED_BACKENDS = ('omniscidb',)


@pytest.mark.parametrize(
    'column',
    [
        'string_col',
        'double_col',
        'date_string_col',
        pytest.param('timestamp_col', marks=pytest.mark.skip(reason='hangs')),
    ],
)
@pytest.mark.xfail_unsupported
def test_distinct_column(backend, alltypes, df, column):
    expr = alltypes[column].distinct()
    result = expr.execute()
    expected = df[column].unique()
    assert set(result) == set(expected)


@pytest.mark.xfail_unsupported
def test_rowid(con, backend):
    t = con.table('functional_alltypes')
    result = t[t.rowid()].execute()
    first_value = 0 if backend.name() in ROWID_ZERO_INDEXED_BACKENDS else 1
    expected = pd.Series(
        range(first_value, first_value + len(result)),
        dtype=np.int64,
        name='rowid',
    )
    pd.testing.assert_series_equal(result.iloc[:, 0], expected)


@pytest.mark.xfail_unsupported
def test_named_rowid(con, backend):
    t = con.table('functional_alltypes')
    result = t[t.rowid().name('number')].execute()
    first_value = 0 if backend.name() in ROWID_ZERO_INDEXED_BACKENDS else 1
    expected = pd.Series(
        range(first_value, first_value + len(result)),
        dtype=np.int64,
        name='number',
    )
    pd.testing.assert_series_equal(result.iloc[:, 0], expected)



from __future__ import annotations

from collections import OrderedDict

import pandas as pd
from clickhouse_driver.client import Client as _DriverClient

import ibis
import ibis.config
import ibis.expr.schema as sch
from ibis.backends.base.sql import BaseSQLBackend
from ibis.config import options

from .client import ClickhouseDataType, ClickhouseTable, fully_qualified_re
from .compiler import ClickhouseCompiler

_default_compression: str | bool

try:
    import lz4  # noqa: F401

    _default_compression = 'lz4'
except ImportError:
    _default_compression = False


class Backend(BaseSQLBackend):
    name = 'clickhouse'
    table_expr_class = ClickhouseTable
    compiler = ClickhouseCompiler

    def connect(
        self,
        host='localhost',
        port=9000,
        database='default',
        user='default',
        password='',
        client_name='ibis',
        compression=_default_compression,
    ):
        """Create an ClickhouseClient for use with Ibis.

        Parameters
        ----------
        host : str, optional
            Host name of the clickhouse server
        port : int, optional
            Clickhouse server's  port
        database : str, optional
            Default database when executing queries
        user : str, optional
            User to authenticate with
        password : str, optional
            Password to authenticate with
        client_name: str, optional
            This will appear in clickhouse server logs
        compression: str, optional
            Weather or not to use compression.
            Default is lz4 if installed else False.
            Possible choices: lz4, lz4hc, quicklz, zstd, True, False
            True is equivalent to 'lz4'.

        Examples
        --------
        >>> import ibis
        >>> import os
        >>> clickhouse_host = os.environ.get('IBIS_TEST_CLICKHOUSE_HOST',
        ...                                  'localhost')
        >>> clickhouse_port = int(os.environ.get('IBIS_TEST_CLICKHOUSE_PORT',
        ...                                      9000))
        >>> client = ibis.clickhouse.connect(
        ...     host=clickhouse_host,
        ...     port=clickhouse_port
        ... )
        >>> client  # doctest: +ELLIPSIS
        <ibis.clickhouse.client.ClickhouseClient object at 0x...>

        Returns
        -------
        ClickhouseClient
        """
        new_backend = self.__class__()
        new_backend.con = _DriverClient(
            host=host,
            port=port,
            database=database,
            user=user,
            password=password,
            client_name=client_name,
            compression=compression,
        )
        return new_backend

    def register_options(self):
        ibis.config.register_option(
            'temp_db',
            '__ibis_tmp',
            'Database to use for temporary tables, views. functions, etc.',
        )

    @property
    def version(self) -> str:
        self.con.connection.force_connect()
        try:
            info = self.con.connection.server_info
        except Exception:
            self.con.connection.disconnect()
            raise

        return f'{info.version_major}.{info.version_minor}.{info.revision}'

    @property
    def current_database(self):
        return self.con.connection.database

    def list_databases(self, like=None):
        data, schema = self.raw_sql('SELECT name FROM system.databases')
        databases = list(data[0])
        return self._filter_with_like(databases, like)

    def list_tables(self, like=None, database=None):
        data, schema = self.raw_sql('SHOW TABLES')
        databases = list(data[0])
        return self._filter_with_like(databases, like)

    def raw_sql(self, query: str, external_tables={}):
        external_tables_list = []
        for name, df in external_tables.items():
            if not isinstance(df, pd.DataFrame):
                raise TypeError(
                    'External table is not an instance of pandas ' 'dataframe'
                )
            schema = sch.infer(df)
            external_tables_list.append(
                {
                    'name': name,
                    'data': df.to_dict('records'),
                    'structure': list(
                        zip(
                            schema.names,
                            [
                                str(ClickhouseDataType.from_ibis(t))
                                for t in schema.types
                            ],
                        )
                    ),
                }
            )

        ibis.util.log(query)
        return self.con.execute(
            query,
            columnar=True,
            with_column_types=True,
            external_tables=external_tables_list,
        )

    def ast_schema(self, query_ast, external_tables={}):
        # Allowing signature to accept `external_tables`
        return super().ast_schema(query_ast)

    def fetch_from_cursor(self, cursor, schema):
        data, columns = cursor
        if not len(data):
            # handle empty resultset
            return pd.DataFrame([], columns=schema.names)

        df = pd.DataFrame.from_dict(OrderedDict(zip(schema.names, data)))
        return schema.apply_to(df)

    def close(self):
        """Close Clickhouse connection and drop any temporary objects"""
        self.con.disconnect()

    def _fully_qualified_name(self, name, database):
        if fully_qualified_re.search(name):
            return name

        database = database or self.current_database
        return f'{database}.`{name}`'

    def get_schema(self, table_name, database=None):
        """
        Return a Schema object for the indicated table and database

        Parameters
        ----------
        table_name : string
          May be fully qualified
        database : string, default None

        Returns
        -------
        schema : ibis Schema
        """
        qualified_name = self._fully_qualified_name(table_name, database)
        query = f'DESC {qualified_name}'
        data, columns = self.raw_sql(query)
        return sch.schema(
            data[0], list(map(ClickhouseDataType.parse, data[1]))
        )

    def set_options(self, options):
        self.con.set_options(options)

    def reset_options(self):
        # Must nuke all cursors
        raise NotImplementedError

    def _ensure_temp_db_exists(self):
        name = (options.clickhouse.temp_db,)
        if name not in self.list_databases():
            self.create_database(name, force=True)

    def _get_schema_using_query(self, query, **kwargs):
        data, columns = self.raw_sql(query, **kwargs)
        colnames, typenames = zip(*columns)
        coltypes = list(map(ClickhouseDataType.parse, typenames))
        return sch.schema(colnames, coltypes)

    def _table_command(self, cmd, name, database=None):
        qualified_name = self._fully_qualified_name(name, database)
        return f'{cmd} {qualified_name}'



from collections import OrderedDict

import ibis
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis.tests.util import assert_pickle_roundtrip


def test_struct_operations():
    value = OrderedDict(
        [
            ('a', 1),
            ('b', list('abc')),
            ('c', OrderedDict([('foo', [1.0, 2.0])])),
        ]
    )
    expr = ibis.literal(value)
    assert isinstance(expr, ir.StructValue)
    assert isinstance(expr['b'], ir.ArrayValue)
    assert isinstance(expr['a'].op(), ops.StructField)


def test_struct_field_dir():
    t = ibis.table([('struct_col', 'struct<my_field: string>')])
    assert 'struct_col' in dir(t)
    assert 'my_field' in dir(t.struct_col)


def test_struct_pickle():
    struct_scalar_expr = ibis.literal(
        OrderedDict([("fruit", "pear"), ("weight", 0)])
    )

    assert_pickle_roundtrip(struct_scalar_expr)



import numpy as np
import pandas as pd
import pandas.testing as tm
import pyspark
import pytest
from pytest import param

import ibis

if pyspark.__version__ < '3.0.0':
    null_representation = None
else:
    null_representation = np.nan


def test_array_length(client):
    table = client.table('array_table')

    result = table.mutate(length=table.array_int.length()).compile()

    expected = table.compile().toPandas()
    expected['length'] = (
        expected['array_int'].map(lambda a: len(a)).astype('int32')
    )
    tm.assert_frame_equal(result.toPandas(), expected)


def test_array_length_scalar(client):
    raw_value = [1, 2, 3]
    value = ibis.literal(raw_value)
    expr = value.length()
    result = client.execute(expr)
    expected = len(raw_value)
    assert result == expected


@pytest.mark.parametrize(
    ['start', 'stop'],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        # negative slices are not supported
        param(
            -3,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            None,
            -3,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            -3,
            -1,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
    ],
)
def test_array_slice(client, start, stop):
    table = client.table('array_table')

    result = table.mutate(sliced=table.array_int[start:stop]).compile()

    expected = table.compile().toPandas()
    expected['sliced'] = expected['array_int'].map(lambda a: a[start:stop])
    tm.assert_frame_equal(result.toPandas(), expected)


@pytest.mark.parametrize(
    ['start', 'stop'],
    [
        (1, 3),
        (1, 1),
        (2, 3),
        (2, 5),
        (None, 3),
        (None, None),
        (3, None),
        # negative slices are not supported
        param(
            -3,
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            None,
            -3,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
        param(
            -3,
            -1,
            marks=pytest.mark.xfail(
                raises=ValueError, reason='Negative slicing not supported'
            ),
        ),
    ],
)
def test_array_slice_scalar(client, start, stop):
    raw_value = [-11, 42, 10]
    value = ibis.literal(raw_value)
    expr = value[start:stop]
    result = client.execute(expr)
    expected = raw_value[start:stop]
    assert result == expected


@pytest.mark.parametrize('index', [1, 3, 4, 11, -11])
def test_array_index(client, index):
    table = client.table('array_table')
    expr = table[table.array_int[index].name('indexed')]
    result = expr.execute()

    df = table.compile().toPandas()
    expected = pd.DataFrame(
        {
            'indexed': df.array_int.apply(
                lambda x: x[index]
                if -len(x) <= index < len(x)
                else null_representation
            )
        }
    )
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('index', [1, 3, 4, 11])
def test_array_index_scalar(client, index):
    raw_value = [-10, 1, 2, 42]
    value = ibis.literal(raw_value)
    expr = value[index]
    result = client.execute(expr)
    expected = (
        raw_value[index] if index < len(raw_value) else null_representation
    )
    assert result == expected or (np.isnan(result) and np.isnan(expected))


@pytest.mark.parametrize('op', [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat(client, op):
    table = client.table('array_table')
    x = table.array_int.cast('array<string>')
    y = table.array_str
    expr = op(x, y)
    result = expr.execute()

    df = table.compile().toPandas()
    expected = op(
        df.array_int.apply(lambda x: list(map(str, x))), df.array_str
    ).rename('tmp')
    tm.assert_series_equal(result, expected)


@pytest.mark.parametrize('op', [lambda x, y: x + y, lambda x, y: y + x])
def test_array_concat_scalar(client, op):
    raw_left = [1, 2, 3]
    raw_right = [3, 4]
    left = ibis.literal(raw_left)
    right = ibis.literal(raw_right)
    expr = op(left, right)
    result = client.execute(expr)
    assert result == op(raw_left, raw_right)


@pytest.mark.parametrize('n', [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize('mul', [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat(client, n, mul):
    table = client.table('array_table')

    expr = table.projection([mul(table.array_int, n).name('repeated')])
    result = expr.execute()

    df = table.compile().toPandas()
    expected = pd.DataFrame({'repeated': df.array_int * n})
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize('n', [1, 3, 4, 7, -2])  # negative returns empty list
@pytest.mark.parametrize('mul', [lambda x, n: x * n, lambda x, n: n * x])
def test_array_repeat_scalar(client, n, mul):
    raw_array = [1, 2]
    array = ibis.literal(raw_array)
    expr = mul(array, n)
    result = client.execute(expr)
    expected = mul(raw_array, n)
    assert result == expected


def test_array_collect(client):
    table = client.table('array_table')
    expr = table.group_by(table.key).aggregate(
        collected=table.array_int.collect()
    )
    result = expr.execute().sort_values('key').reset_index(drop=True)

    df = table.compile().toPandas()
    expected = (
        df.groupby('key')
        .array_int.apply(list)
        .reset_index()
        .rename(columns={'array_int': 'collected'})
    )
    tm.assert_frame_equal(result, expected)



try:
    import geoalchemy2  # noqa F401
    import geoalchemy2.shape  # noqa F401
    import geopandas  # noqa F401
except ImportError:
    geospatial_supported = False
else:
    geospatial_supported = True



import os
from pathlib import Path

import numpy as np
import pandas as pd
import pytest


@pytest.fixture(scope='session')
def data_directory() -> Path:
    """
    Fixture that returns the test data directory.

    Returns
    -------
    Path
        Test data directory
    """
    root = Path(__file__).absolute().parent.parent.parent

    default = root / 'ci' / 'ibis-testing-data'
    datadir = os.environ.get('IBIS_TEST_DATA_DIRECTORY', default)
    datadir = Path(datadir)

    return datadir


@pytest.fixture
def file_backends_data():
    # basic time/ticker frame
    rng = pd.date_range('20170101', periods=10, freq='D')
    tickers = ['GOOGL', 'FB', 'APPL', 'NFLX', 'AMZN']
    df = pd.DataFrame(
        {
            'time': np.repeat(rng, len(tickers)),
            'ticker': np.tile(tickers, len(rng)),
        }
    )
    opens = df.assign(open=np.random.randn(len(df)))
    closes = df.assign(close=np.random.randn(len(df)))
    return {'open': opens, 'close': closes}



"""Top level APIs for defining vectorized UDFs.

Warning: This is an experimental module and API here can change without notice.

DO NOT USE DIRECTLY.
"""

import functools

import ibis.expr.datatypes as dt
import ibis.udf.validate as v
from ibis.expr.operations import (
    AnalyticVectorizedUDF,
    ElementWiseVectorizedUDF,
    ReductionVectorizedUDF,
)
from ibis.expr.schema import (
    coerce_to_dataframe,
    coerce_to_np_array,
    coerce_to_series,
    coerce_to_tuple,
)


class UserDefinedFunction:
    """Class representing a user defined function.

    This class Implements __call__ that returns an ibis expr for the UDF.
    """

    def __init__(self, func, func_type, input_type, output_type):
        v.validate_input_type(input_type, func)
        v.validate_output_type(output_type)

        self.func = func
        self.func_type = func_type
        self.input_type = list(map(dt.dtype, input_type))
        self.output_type = dt.dtype(output_type)
        self.coercion_fn = self._get_coercion_function()

    def _get_coercion_function(self):
        """Return the appropriate function to coerce the result of the UDF,
        according to the func type and output type of the UDF."""
        if isinstance(self.output_type, dt.Struct):
            # Case 1: Struct output, non-reduction UDF -> coerce to DataFrame
            if (
                self.func_type is ElementWiseVectorizedUDF
                or self.func_type is AnalyticVectorizedUDF
            ):
                return coerce_to_dataframe
            else:
                # Case 2: Struct output, reduction UDF -> coerce to tuple
                return coerce_to_tuple
        # Case 3: Vector output, non-reduction UDF -> coerce to Series
        elif (
            self.func_type is ElementWiseVectorizedUDF
            or self.func_type is AnalyticVectorizedUDF
        ):
            return coerce_to_series
        # Case 4: Array output type, reduction UDF -> coerce to np.ndarray
        elif isinstance(self.output_type, dt.Array):
            return coerce_to_np_array
        else:
            # Case 5: Default, do nothing (e.g. reduction UDF returning
            # len-0 value such as a single integer or float).
            return None

    def __call__(self, *args, **kwargs):
        # kwargs cannot be part of the node object because it can contain
        # unhashable object, e.g., list.
        # Here, we keep the node hashable by creating a closure that contains
        # kwargs.
        @functools.wraps(self.func)
        def func(*args):
            # If cols are pd.Series, then we save and restore the index.
            saved_index = getattr(args[0], 'index', None)
            result = self.func(*args, **kwargs)
            if self.coercion_fn:
                # coercion function signature must take result, output type,
                # and optionally the index
                result = self.coercion_fn(
                    result, self.output_type, saved_index
                )
            return result

        op = self.func_type(
            func=func,
            args=args,
            input_type=self.input_type,
            output_type=self.output_type,
        )

        return op.to_expr()


def _udf_decorator(node_type, input_type, output_type):
    def wrapper(func):
        return UserDefinedFunction(func, node_type, input_type, output_type)

    return wrapper


def analytic(input_type, output_type):
    """Define an *analytic* user-defined function that takes N
    pandas Series or scalar values as inputs and produces N rows of output.

    Parameters
    ----------
    input_type : List[ibis.expr.datatypes.DataType]
        A list of the types found in :mod:`~ibis.expr.datatypes`. The
        length of this list must match the number of arguments to the
        function. Variadic arguments are not yet supported.
    output_type : ibis.expr.datatypes.DataType
        The return type of the function.

    Examples
    --------
    >>> import ibis
    >>> import ibis.expr.datatypes as dt
    >>> from ibis.udf.vectorized import analytic
    >>> @analytic(input_type=[dt.double], output_type=dt.double)
    ... def zscore(series):  # note the use of aggregate functions
    ...     return (series - series.mean()) / series.std()

    Define and use an UDF with multiple return columns:

    >>> @analytic(
    ...     input_type=[dt.double],
    ...     output_type=dt.Struct(['demean', 'zscore'], [dt.double, dt.double])
    ... )
    ... def demean_and_zscore(v):
    ...     mean = v.mean()
    ...     std = v.std()
    ...     return v - mean, (v - mean) / std
    >>>
    >>> win = ibis.window(preceding=None, following=None, group_by='key')
    >>> # add two columns "demean" and "zscore"
    >>> table = table.mutate(
    ...     demean_and_zscore(table['v']).over(win).destructure()
    ... )
    """
    return _udf_decorator(AnalyticVectorizedUDF, input_type, output_type)


def elementwise(input_type, output_type):
    """Define a UDF (user-defined function) that operates element wise on a
    Pandas Series.

    Parameters
    ----------
    input_type : List[ibis.expr.datatypes.DataType]
        A list of the types found in :mod:`~ibis.expr.datatypes`. The
        length of this list must match the number of arguments to the
        function. Variadic arguments are not yet supported.
    output_type : ibis.expr.datatypes.DataType
        The return type of the function.

    Examples
    --------
    >>> import ibis
    >>> import ibis.expr.datatypes as dt
    >>> from ibis.udf.vectorized import elementwise
    >>> @elementwise(input_type=[dt.string], output_type=dt.int64)
    ... def my_string_length(series):
    ...     return series.str.len() * 2

    Define an UDF with non-column parameters:

    >>> @elementwise(input_type=[dt.string], output_type=dt.int64)
    ... def my_string_length(series, *, times):
    ...     return series.str.len() * times

    Define and use an UDF with multiple return columns:

    >>> @elementwise(
    ...     input_type=[dt.string],
    ...     output_type=dt.Struct(['year', 'monthday'], [dt.string, dt.string])
    ... )
    ... def year_monthday(date):
    ...     return date.str.slice(0, 4), date.str.slice(4, 8)
    >>>
    >>> # add two columns "year" and "monthday"
    >>> table = table.mutate(year_monthday(table['date']).destructure())
    """
    return _udf_decorator(ElementWiseVectorizedUDF, input_type, output_type)


def reduction(input_type, output_type):
    """Define a user-defined reduction function that takes N pandas Series
    or scalar values as inputs and produces one row of output.

    Parameters
    ----------
    input_type : List[ibis.expr.datatypes.DataType]
        A list of the types found in :mod:`~ibis.expr.datatypes`. The
        length of this list must match the number of arguments to the
        function. Variadic arguments are not yet supported.
    output_type : ibis.expr.datatypes.DataType
        The return type of the function.

    Examples
    --------
    >>> import ibis
    >>> import ibis.expr.datatypes as dt
    >>> from ibis.udf.vectorized import reduction
    >>> @reduction(input_type=[dt.string], output_type=dt.int64)
    ... def my_string_length_agg(series, **kwargs):
    ...     return (series.str.len() * 2).sum()

    Define and use an UDF with multiple return columns:

    >>> @reduction(
    ...     input_type=[dt.double],
    ...     output_type=dt.Struct(['mean', 'std'], [dt.double, dt.double])
    ... )
    ... def mean_and_std(v):
    ...     return v.mean(), v.std()
    >>>
    >>> # create aggregation columns "mean" and "std"
    >>> table = table.groupby('key').aggregate(
    ...     mean_and_std(table['v']).destructure()
    ... )
    """
    return _udf_decorator(ReductionVectorizedUDF, input_type, output_type)
